from sentio_prober_control.Sentio.Enumerations import SoftwareFence
from sentio_prober_control.Sentio.Response import Response
from sentio_prober_control.Sentio.CommandGroups.ModuleCommandGroupBase import ModuleCommandGroupBase


class ServiceCommandGroup(ModuleCommandGroupBase):
    """A command group for accessing service module functions."""

    def __init__(self, comm) -> None:
        super().__init__(comm, "service")


    def set_compensation_mode(self, status: bool) -> None:
        """Turn chuck compensation on or off.

        Args:
            status: True to turn on, False to turn off.

        Returns:
            A Response object.
        """

        self.comm.send(f"service:chuck_compensation {status}")
        Response.check_resp(self.comm.read_line())


    def set_software_fence(self, fence: SoftwareFence) -> None:
        """Set the software fence.

        The software fence is a virtual fence that is used to
        limit the movement of the chuck. The fences purpose
        is to prevent the chuck from moving into physical
        obstructions such as the machine casing or other
        hardware.

        Args:
            fence: The type of fence to use.

        Returns:
            A Response object.
        """

        self.comm.send(f"service:chuck_fence {fence.toSentioArg()}")
        Response.check_resp(self.comm.read_line())



from abc import ABC


class CommunicatorBase(ABC):
    """Base class for all communicators.

    Communicators ore objects that implement different communication protocols
    for talking to the probe station.
    """

    _verbose = False

    def connect(self):
        """Connect to the probe station.

        Must be implemented by the derived class.
        """
        raise NotImplementedError("CommunicatorBase.connect is not implemented!")

    def disconnect(self):
        """Disconnect from the probe station.

        Must be implemented by the derived class.
        """
        raise NotImplementedError("CommunicatorBase.disconnect is not implemented!")

    def send(self, msg: str):
        """Send a command to the probe station.

        Must be implemented by the derived class.
        """
        raise NotImplementedError("CommunicatorBase.send is not implemented!")

    def read_line(self):
        """Read a line from the probe station.

        Must be implemented by the derived class.
        """
        raise NotImplementedError("CommunicatorBase.read_line is not implemented!")



from typing import Tuple

from sentio_prober_control.Sentio.Enumerations import AutoFocusAlgorithm, CameraMountPoint, DefaultPattern
from sentio_prober_control.Sentio.Response import Response
from sentio_prober_control.Sentio.CommandGroups.CommandGroupBase import CommandGroupBase


class VisionCameraCommandGroup(CommandGroupBase):
    """A command group for setting camera parameters.

    You are not meant to instantiate this class directly. Access it via the camera attribute
    of the vision attribute of the [SentioProber](SentioProber.md) class.
    """

    def __init__(self, comm) -> None:
        super().__init__(comm)


    def set_light(self, mp: CameraMountPoint, value: int) -> Response:
        """Set intensity of the light.

        Args:
            mp: The mount point of the camera.
            value: The intensity of the light.
        """

        self.comm.send(f"vis:set_prop light, {mp.toSentioAbbr()}, {value}")
        return Response.check_resp(self.comm.read_line())


    def set_exposure(self, mp: CameraMountPoint, value: int) -> Response:
        """Set exposure time of the camera.

        Args:
            mp: The mount point of the camera.
            value: The exposure time in microseconds.

        Returns:
            A Response object.
        """

        self.comm.send(f"vis:set_prop exposure, {mp.toSentioAbbr()}, {value}")
        return Response.check_resp(self.comm.read_line())


    def set_gain(self, mp: CameraMountPoint, value: float) -> Response:
        """Set gain of the camera.

        Args:
            mp: The mount point of the camera.
            value: The gain value.
        """

        self.comm.send(f"vis:set_prop gain, {mp.toSentioAbbr()}, {value}")
        return Response.check_resp(self.comm.read_line())


    def get_light(self, mp: CameraMountPoint) -> float:
        """Get light intensity.

        Args:
            mp: The mount point of the camera.
        """

        self.comm.send(f"vis:get_prop light, {mp.toSentioAbbr()}")
        resp = Response.check_resp(self.comm.read_line())
        return float(resp.message())


    def get_exposure(self, mp: CameraMountPoint) -> float:
        """Get exposure time.

        Args:
            mp: The mount point of the camera.

        Returns:
            The exposure time in microseconds.
        """

        self.comm.send(f"vis:get_prop exposure, {mp.toSentioAbbr()}")
        resp = Response.check_resp(self.comm.read_line())
        return float(resp.message())


    def get_focus_value(self, mp: CameraMountPoint, alg: AutoFocusAlgorithm) -> float:
        """Get the focus value of the camera.

        This function wraps the vis:get_focus_value remote command.

        Args:
            mp: The mount point of the camera.
            alg: The autofocus algorithm to use.

        Returns:
            The focus value.
        """

        self.comm.send(
            f"vis:get_focus_value {mp.toSentioAbbr()}, {alg.toSentioAbbr()}"
        )
        resp = Response.check_resp(self.comm.read_line())
        return float(resp.message())


    def get_gain(self, mp: CameraMountPoint) -> float:
        """Get gain of the camera.

        Args:
            mp: The mount point of the camera.

        Returns:
            The gain value.
        """

        self.comm.send(f"vis:get_prop gain, {mp.toSentioAbbr()}")
        resp = Response.check_resp(self.comm.read_line())
        return float(resp.message())


    def get_calib(self, mp: CameraMountPoint) -> Tuple[float, float]:
        """Get the calibration data of the camera.

        Camera calibration data consists of the width and height of
        an image pixel in micrometers.

        Args:
            mp: The mount point of the camera.

        Returns:
            width: Get the width of a camera pixel in micrometers.
            height: Get height of a camera pixel in micrometers.
        """

        self.comm.send(f"vis:get_prop calib, {mp.toSentioAbbr()}")
        resp = Response.check_resp(self.comm.read_line())
        tok = resp.message().split(",")
        return float(tok[0]), float(tok[1])


    def get_image_size(self, mp: CameraMountPoint) -> Tuple[int, int]:
        """Get size of the image.

        Args:
            mp: The mount point of the camera.

        Returns:
            width: The width of the image in pixels.
            height: The height of the image in pixels.
        """

        self.comm.send(f"vis:get_prop image_size, {mp.toSentioAbbr()}")
        resp = Response.check_resp(self.comm.read_line())
        tok = resp.message().split(",")
        return int(tok[0]), int(tok[1])


    def is_pattern_trained(self, mp: CameraMountPoint, pat) -> bool:
        """Check if a pattern is trained.

        This command wraps the vis:pattern:is_trained remote command.

        Args:
            mp: The mount point of the camera.

        Returns:
            True if the pattern is trained, False otherwise.
        """

        if isinstance(pat, DefaultPattern):
            pattern_name = pat.toSentioAbbr()
        else:
            pattern_name = pat

        self.comm.send(f"vis:pattern:is_trained {mp.toSentioAbbr()}, {pattern_name}")
        resp = Response.check_resp(self.comm.read_line())
        is_trained = resp.message() == "1"
        return is_trained



""" @private """
import ctypes
import re
from ctypes import c_byte, c_char_p, c_int, c_ushort, c_void_p, c_wchar_p


class GpibNiDriver:
    """A python wrapper for the NI gpib-32.dll.

    This class is used internally only.

    @private
    """

    def __init__(self):
        # load gpib dll
        gpibDll = ctypes.WinDLL("ni4882.dll")

        # gpib init functions
        self._gpibSendIfc = gpibDll.SendIFC
        self._gpibSendIfc.argtypes = [c_int]

        self._gpibThreadIbsta = gpibDll.ThreadIbsta
        self._gpibThreadIbsta.restype = c_int

        self._gpibThreadIberr = gpibDll.ThreadIberr
        self._gpibThreadIberr.restype = c_int

        self._gpibThreadIbcnt = gpibDll.ThreadIbcnt
        self._gpibThreadIbcnt.restype = c_int

        self._gpibIbfind = gpibDll.ibfindW
        self._gpibIbfind.restype = c_int
        self._gpibIbfind.argtypes = [c_wchar_p]

        self._gpibIbsic = gpibDll.ibsic
        self._gpibIbsic.restype = c_int
        self._gpibIbsic.argtypes = [c_int]

        # gpib send
        self._gpibSendImpl = gpibDll.Send
        self._gpibSendImpl.argtypes = [c_int, c_ushort, c_char_p, c_int, c_int]

        # gpib receive
        self._gpibReceiveImpl = gpibDll.Receive
        self._gpibReceiveImpl.argtypes = [c_int, c_ushort, c_void_p, c_int, c_int]

    def connect(self, board_name: str, address: int):
        # check wether the board exists
        board_handle = self._gpibIbfind(board_name)
        if board_handle < 0:
            raise Exception(f'Board "{board_name}" does not exist!')

        # init
        numbers = re.findall("[0-9]+", board_name)
        self._board = int(numbers[0])

        self._gpibSendIfc(self._board)
        ibsta, iberr, ibcnt, ibcntl = self.get_globals()

        if iberr != 0:
            raise Exception(f"Connect failed: Gpib error code {iberr}")

        self._addr = address
        self._max_len = 1000

    def get_globals(self):
        ibsta = self._gpibThreadIbsta()
        iberr = self._gpibThreadIberr()
        ibcnt = 0
        ibcntl = self._gpibThreadIbcnt()

        return ibsta, iberr, ibcnt, ibcntl

    def send(self, str):
        self._gpibSendImpl(self._board, self._addr, str.encode("utf-8"), len(str), 1)

        iberr = self._gpibThreadIberr()
        if iberr != 0:
            raise Exception(f'Sending "{str}" failed. (GPIB error code {iberr})')

    def receive(self):
        byteArr = (c_byte * self._max_len)()
        self._gpibReceiveImpl(self._board, self._addr, byteArr, self._max_len, 0)
        ibsta, iberr, ibcnt, ibcntl = self.get_globals()

        if iberr != 0:
            raise Exception(f"Receive failed: Gpib error code {iberr}")

        return bytearray(byteArr[:ibcntl]).decode("utf-8")



from typing import Tuple

from sentio_prober_control.Sentio.Enumerations import AxisOrient, ColorScheme, DieNumber, StatusBits
from sentio_prober_control.Sentio.ProberBase import ProberException
from sentio_prober_control.Sentio.Response import Response
from sentio_prober_control.Sentio.CommandGroups.ModuleCommandGroupBase import ModuleCommandGroupBase
from sentio_prober_control.Sentio.CommandGroups.WafermapBinsCommandGroup import WafermapBinsCommandGroup
from sentio_prober_control.Sentio.CommandGroups.WafermapCompensationCommandGroup import WafermapCompensationCommandGroup
from sentio_prober_control.Sentio.CommandGroups.WafermapDieCommandGroup import WafermapDieCommandGroup
from sentio_prober_control.Sentio.CommandGroups.WafermapPathCommandGroup import WafermapPathCommandGroup
from sentio_prober_control.Sentio.CommandGroups.WafermapPoiCommandGroup import WafermapPoiCommandGroup
from sentio_prober_control.Sentio.CommandGroups.WafermapSubsiteCommandGroup import WafermapSubsiteGroup


class WafermapCommandGroup(ModuleCommandGroupBase):
    """This class represents the SENTIO command group for wafermap related commands.
    You are not meant to instantiate objects of this class directly! This class
    is instantiated by the prober implementation.

    Examples:

    ```py
    from sentio_prober_control.Sentio.ProberSentio import SentioProber
    from sentio_prober_control.Communication.CommunicatorTcpIp import CommunicatorTcpIp

    prober = SentioProber(CommunicatorTcpIp.create("127.0.0.1:35555"))
    prober.map.create(200)
    ```

    Attributes:
        bins (WafermapBinsCommandGroup): Command group with functions for setting up binning tables.
        die (WafermapDieCommandGroup): A group to set up specific dies on the wafermap (add/remove them).
        path (WafermapPathCommandGroup): A group to set up test paths.
        poi (WafermapPoiCommandGroup): A group to set up points of interest.
        subsites (WafermapSubsiteGroup): A group to set up subsites.
    """

    def __init__(self, comm) -> None:
        super().__init__(comm, "map")

        self.__end_of_route: bool = False

        self.bins: WafermapBinsCommandGroup = WafermapBinsCommandGroup(comm)
        self.compensation: WafermapCompensationCommandGroup = WafermapCompensationCommandGroup(comm)
        self.die: WafermapDieCommandGroup = WafermapDieCommandGroup(comm)
        self.path: WafermapPathCommandGroup = WafermapPathCommandGroup(comm)
        self.poi: WafermapPoiCommandGroup = WafermapPoiCommandGroup(comm)
        self.subsites: WafermapSubsiteGroup = WafermapSubsiteGroup(comm, self)


    def bin_step_next_die(self, bin_value: int, site: int | None = None) -> Tuple[int, int, int]:
        """Bin the current die and step to the naxt die.

        This command wraps the "map:bin_step_next_die" remote command.

        Args:
            bin_value (int): The bin value to set for the current die.
            site (int): The subsite index of the next die to move to. Defaults to None. By default the current subsite is retained.

        Returns:
            columns (int): The column index of the die after stepping
            rows (int): The row index of the die after stepping
            subsite (int): The subsite index of the die after stepping
        """

        # 2021-09-17: bugfix: when no site is given current site must be retained
        if site is None:
            self.comm.send(f"map:bin_step_next_die {bin_value}")
        else:
            self.comm.send(f"map:bin_step_next_die {bin_value}, {site}")

        resp = Response.parse_resp(self.comm.read_line())
        self.__end_of_route = (
            resp.status() & StatusBits.EndOfRoute
        ) == StatusBits.EndOfRoute

        # i.e. Stepping while at the end of the route
        if not resp.ok():
            raise ProberException(resp.message(), resp.errc())

        tok = resp.message().split(",")
        return int(tok[0]), int(tok[1]), int(tok[2])


    def create(self, diameter: float) -> None:
        """Create a new round wafer map.

        Wraps Sentios "map:create" remote command.

        Args:
            diameter (float): The diameter of the wafer.
        """

        self.comm.send(f"map:create {diameter}")
        Response.check_resp(self.comm.read_line())


    def create_rect(self, cols: int, rows: int) -> None:
        """Create a new rectangular wafer map.

        Wraps Sentios "map:create_rect" remote command.

        Args:
            cols (int): The number of columns.
            rows (int): The number of rows.
        """

        self.comm.send("map:create_rect {0}, {1}".format(cols, rows))
        Response.check_resp(self.comm.read_line())


    def die_reference_is_set(self) -> bool:
        """Returns true if the die reference offset is set.

        In simple terms the die reference offset is the distance of site 0 from
        the lower left corner of the die. This position defines where
        the die is located with respect to the die in particular and
        the wafer map grid in extension.

        When the chuck is at site 0 and the probe card is properly in contact
        at the home position, then the die reference offset is the distance between
        a user defined reference tip of the probe card and the lower left corner
        of the die.

        Without a properly trained die reference position SENTIO is unable
        to accurately deterine the die coordinates from a chuck position.
        Most of the time this is not a problem.

        When a die reference is trained Sentio will use the die reference offset
        to calculate the die coordinates of the die which is under the reference
        tip of the probe card. The reference tip can be any tip of the probe card.

        Wraps Sentios "map:get_prop die_reference_is_set" remote command.

        Returns:
            True if the die reference offset is set.
        """
        self.comm.send("map:get_prop die_reference_is_set")
        resp = Response.check_resp(self.comm.read_line())
        return resp.message().lower() == "true"


    def end_of_route(self) -> bool:
        """Returns True if the last stepping command reached the end of the route.

        The end of route flag is checked and set by the stepping commands.
        They will evaluate and parse the end of route status from the remote
        command response and internally set this flag.

        Calling the function is only usefull after a step command was issued.

        Returns:
            True if the last step command reached the end of the route.
        """
        return self.__end_of_route


    def get_axis_orient(self) -> AxisOrient:
        """Get axis orientation of the wafer map.

        Wraps Sentios "map:get_axis_orient" remote command.

        Returns:
            The axis orientation.
        """

        self.comm.send("map:get_axis_orient")
        resp = Response.check_resp(self.comm.read_line())

        if resp.message().upper() == "UL":
            return AxisOrient.UpLeft

        if resp.message().upper() == "UR":
            return AxisOrient.UpRight

        if resp.message().upper() == "DR":
            return AxisOrient.DownRight

        if resp.message().upper() == "DL":
            return AxisOrient.DownLeft
        
        raise ProberException(f"Unknown axis orientation: {resp.message()}")


    def get_diameter(self) -> float:
        """Get diameter of the wafer map im millimeter.

        Wraps Sentios "map:get_diameter" remote command.

        Returns:
            The diameter of the wafer map in millimeter.
        """
        self.comm.send("map:get_diameter")
        resp = Response.check_resp(self.comm.read_line())

        dia = int(resp.message())
        return dia


    def get_die_reference(self) -> Tuple[float, float]:
        """Get the die reference offset.

        In simple terms the die reference offset is the distance of site 0 from
        the lower left corner of the die. This position defines where
        the die is located with respect to the die in particular and
        the wafer map grid in extension.

        When the chuck is at site 0 and the probe card is properly in contact
        at the home position, then the die reference offset is the distance between
        a user defined reference tip of the probe card and the lower left corner
        of the die.

        Without a properly trained die reference position SENTIO is unable
        to accurately deterine the die coordinates from a chuck position.
        Most of the time this is not a problem.

        When a die reference is trained Sentio will use the die reference offset
        to calculate the die coordinates of the die which is under the reference
        tip of the probe card. The reference tip can be any tip of the probe card.

        Wraps SENTIO's "map:get_prop die_reference" remote command.

        Returns:
            A tuple with the x and y offset in micrometer.
        """
        self.comm.send("map:get_prop die_reference")
        resp = Response.check_resp(self.comm.read_line())

        tok = resp.message().split(",")
        return float(tok[0]), float(tok[1])


    def get_die_seq(self) -> int:
        """Returns the sequence number of the current die.

        The sequence number id the index of the die in the
        current stepping sequence. The stepping sequence
        contains the dies selected for test.
        The sequence number is zero based and is also displayed
        in the wafermap when zoomin into a die.

        Wraps Sentios "map:get_die_seq" remote command.

        Returns:
            The sequence number of the current die.
        """
        self.comm.send("map:get_die_seq")
        resp = Response.check_resp(self.comm.read_line())
        return int(resp.message())  # 0:Result+status, 1:Command ID, 2:Response


    def get_grid_origin(self) -> Tuple[int, int]:
        """Get origin of the wafermap grid.

        Wraps Sentios "map:get_grid_origin" remote command.

        Returns:
            A tuple with the column and row indices of the origin.
        """
        self.comm.send("map:get_grid_origin")
        resp = Response.check_resp(self.comm.read_line())
        tok = resp.message().split(",")

        return int(tok[0]), int(tok[1])


    def get_index_size(self) -> Tuple[float, float]:
        """Return the die size set up in the wafer map.

        Returns:
            A tuple with the die width and height in micrometer.
        """
        self.comm.send("map:get_index_size")
        resp = Response.check_resp(self.comm.read_line())
        tok = resp.message().split(",")
        return float(tok[0]), float(tok[1])


    def get_num_dies(self, selection: DieNumber) -> int:
        """Returns the number of dies in the wafer map.

        Args:
            selection: The selection of dies to count.

        Returns:
            The number of dies.
        """
        switcher = {DieNumber.Present: "Present", DieNumber.Selected: "Selected"}

        what = switcher.get(selection, "Invalid die number selector")

        self.comm.send("map:get_num_dies {0}".format(what))
        resp = Response.check_resp(self.comm.read_line())
        return int(resp.message())


    def get_street_size(self) -> Tuple[int, int]:
        """Returns the street size set up in the wafer map.

        In SENTIO the street size is merely a visual aid. It is not used for
        any calculations. The only purpose is rendering the diese in a more
        realistic manner.

        Returns:
            A tuple with the street width and height in micrometer.
        """
        self.comm.send("map:get_street_size")
        resp = Response.check_resp(self.comm.read_line())
        tok = resp.message().split(",")

        return int(tok[0]), int(tok[1])


    def set_axis_orient(self, orient: AxisOrient) -> None:
        """Set the acis orientation of the custom coordinate system.

        SENTIO's internal coordinate system is placing the origin at the center of the wafer and
        assumes the y-axis is pointing upwards and the x-axis is pointing to the right. With
        this command you can change the orientation of the coordinate system.

        Wraps Sentios "map:set_axis_orient" remote command.

        Args:
            orient: The axis orientation.
        """
        self.comm.send(f"map:set_axis_orient {orient.toSentioAbbr()}")
        Response.check_resp(self.comm.read_line())


    def set_color_scheme(self, scheme: ColorScheme) -> None:
        """Set color scheme of the wafermap.

        The wafermap colorscheme determines how dies are colored.
        The only two options are color from bin or color from value.
        The first one will use the binning table to color the die the second
        one will apply a color scale based on a numeric value associated with
        the die. (Each die can have a singe temporary numerical value assigned
        to it.)

        Wraps Sentios "map:set_color_scheme" remote command.

        Args:
            scheme: The color scheme.
        """
        self.comm.send(f"map:set_color_scheme {scheme.toSentioAbbr()}")
        Response.check_resp(self.comm.read_line())


    def set_flat_params(self, angle: float, width: float) -> None:
        """Set the flat parameters of the wafer map.

        Sets the parameters of the flat wafer orientation marker for
        display in the wafermap.

        Args:
            angle: The angle of the flat in degrees.
            width: The width of the flat in micrometer.
        """
        self.comm.send("map:set_flat_params {0}, {1}".format(angle, width))
        Response.check_resp(self.comm.read_line())


    def set_grid_origin(self, x: int, y: int) -> None:
        """Set a user defined grid origin.

        SENTIO's internal coordinate system is placing the origin at the center of the wafer and
        assumes the y-axis is pointing upwards and the x-axis is pointing to the right. You may
        want to use a custom coordinate system for your wafermap. Coordinates with customized
        grid origin are referred to as custom coordinates.

        Wraps Sentios "map:set_grid_origin" remote command.

        Args:
            x: The x coordinate of the grid origin in SENTIO's native grid coordinate system.
            y: The y coordinate of the grid origin in SENTIO's native grid coordinate system.
        """
        self.comm.send(f"map:set_grid_origin {x}, {y}")
        Response.check_resp(self.comm.read_line())


    def set_grid_params(self, ix: float, iy: float, offx: float, offy: float, edge: int) -> None:
        """Set wafermap grid parameters. This function defines the wafermapo grid layout which means setting the
        size of a die. Setting the grid offset and the size of the edge exclusion zone.

        By default SENTIO assumes that the grid is oriented in a way that the center of the wafer
        grid is exactly at the lower left corner of the die located in the center. This would
        therefore also be the position of site 0.

        It is not recommended to set the grid offset to 0,0 because this may pose problems
        when computing the current die index from a chuck position. Due to inaccuracies in the chuck
        motion the computation may end up in the wrong die because with grid offset of 0,0
        this position is at the edge of 4 dies.

        It is recommended to either determine the correct value of the die reference offset or
        to use half the die size as grid offset.

        Wraps Sentios "map:set_grid_params" remote command.

        Args:
            ix: The width of a single die in micrometer.
            iy: The height of a single die in micrometer.
            offx: The x offset of the grid in micrometer.
            offy: The y offset of the grid in micrometer.
            edge: The size of the edge exclusion zone in micrometer.
        """
        self.comm.send(f"map:set_grid_params {ix}, {iy}, {offx}, {offy}, {edge}")
        Response.check_resp(self.comm.read_line())


    def set_home_die(self, x: int, y: int) -> None:
        """ " Sets the home die coordinates in custom coordinates.

        Wraps Sentios "map:set_home_die" remote command.

        Args:
            x: The x coordinate of the home die in custom coordinates.
            y: The y coordinate of the home die in custom coordinates.
        """
        self.comm.send(f"map:set_home_die {x}, {y}")
        Response.check_resp(self.comm.read_line())


    def set_index_size(self, x: float, y: float) -> None:
        """Set the size of a die.

        Wraps Sentios "map:set_index_size" remote command.

        Args:
            x: The width of the die in micrometer.
            y: The height of the die in micrometer.
        """
        self.comm.send("map:set_index_size {0}, {1}".format(x, y))
        Response.check_resp(self.comm.read_line())


    def set_street_size(self, x: float, y: float) -> None:
        """Set size of streetlines.

        In SENTIO the street size is merely a visual aid. It is not used for
        any calculations. The only purpose is rendering the diese in a more
        realistic manner.

        Args:
            x: The width of the street in micrometer.
            y: The height of the street in micrometer.
        """
        self.comm.send("map:set_street_size {0}, {1}".format(x, y))
        Response.check_resp(self.comm.read_line())


    def step_die(self, col: int, row: int, site: int = 0) -> Tuple[int, int, int]:
        """Step to a specific die (or subsite) which is identified by its column,
        row and subsite number.

        When the stepping command is issued on the last die of the route, the
        end of route flag is set and can be queried with the end_of_route member function.

        Wraps Sentios "map:step_die" remote command.

        Args:
            col: The column index of the die.
            row: The row index of the die.
            site: The subsite index of the die. Defaults to 0.

        Returns:
            A tuple with the column, row and site index representing the position after the step command.
        """
        self.comm.send("map:step_die {0}, {1}, {2}".format(col, row, site))
        resp = Response.parse_resp(self.comm.read_line())

        self.__end_of_route = (
            resp.status() & StatusBits.EndOfRoute
        ) == StatusBits.EndOfRoute

        if not resp.ok():
            raise ProberException(resp.message(), resp.errc())

        tok = resp.message().split(",")
        return int(tok[0]), int(tok[1]), int(tok[2])


    def step_die_seq(self, seq: int, site: int) -> Tuple[int, int, int]:
        """Step to a specific die in the stepping sequence.

        Wraps Sentios "map:step_die_seq" remote command.

        Args:
            seq: The sequence number of the die to step to.
            site: The subsite index of the die.

        Returns:
            A tuple with the column, row and site index representing the position after the step command.
        """

        self.comm.send("map:step_die_seq {}, {}".format(seq, site))
        resp = Response.check_resp(self.comm.read_line())
        tok = resp.message().split(",")

        # i.e. Stepping while at the end of the route
        if not resp.ok():
            raise ProberException(resp.message())

        return int(tok[0]), int(tok[1]), int(tok[2])


    def step_first_die(self, site: int | None = None) -> Tuple[int, int, int]:
        """Step to the first die in the stepping sequence.

        Wraps Sentios "map:step_first_die" remote command.

        Args:
            site: The subsite index of the die. If omitted SENTIO will step to the first active site.

        Returns:
            A tuple with the column, row and site index representing the position after the step command.
        """

        if site == None:
            self.comm.send("map:step_first_die")
        else:
            self.comm.send(f"map:step_first_die {site}")

        resp = Response.parse_resp(self.comm.read_line())

        self.__end_of_route = (resp.status() & StatusBits.EndOfRoute) == StatusBits.EndOfRoute

        if not resp.ok():
            raise ProberException(resp.message(), resp.errc())

        tok = resp.message().split(",")

        return int(tok[0]), int(tok[1]), int(tok[2])


    def step_next_die(self, site: int | None = None) -> Tuple[int, int, int]:
        """Step to the next die in the stepping sequence.

        Sentio will stay on the current subsite.

        Returns:
            A tuple with the column, row and site index representing the position after the step command.
        """

        # 2021-09-17: bugfix: when no site is given current site must be retained
        if site is None:
            self.comm.send(f"map:step_next_die")
        else:
            self.comm.send(f"map:step_next_die {site}")

        resp = Response.parse_resp(self.comm.read_line())
        self.__end_of_route = (resp.status() & StatusBits.EndOfRoute) == StatusBits.EndOfRoute

        # i.e. Stepping while at the end of the route
        if not resp.ok():
            raise ProberException(resp.message(), resp.errc())

        tok = resp.message().split(",")
        return int(tok[0]), int(tok[1]), int(tok[2])



from deprecated import deprecated

from sentio_prober_control.Sentio.Enumerations import ExecuteAction
from sentio_prober_control.Sentio.ProberBase import ProberException
from sentio_prober_control.Sentio.Response import Response
from sentio_prober_control.Sentio.CommandGroups.CommandGroupBase import CommandGroupBase


@deprecated("Use VisionCompensationGroup instead")
class WafermapCompensationCommandGroup(CommandGroupBase):

    @deprecated("Use vis.compensation.start_execute(CompensationType.Topography, CompensationMode.Vertical) instead")
    def topography(self, execute: ExecuteAction):
        self.comm.send(f"map:compensation:topography {execute.toSentioAbbr()}")
        resp = Response.check_resp(self.comm.read_line())

        # i.e. Stepping while at the end of the route
        if not resp.ok():
            raise ProberException(resp.message())

        return resp.cmd_id()



from typing import Tuple

from sentio_prober_control.Communication.CommunicatorBase import CommunicatorBase
from sentio_prober_control.Sentio.Enumerations import ProbeSentio, ProbeXYReference, ProbeZReference, ChuckSite
from sentio_prober_control.Sentio.Response import Response
from sentio_prober_control.Sentio.CommandGroups.CommandGroupBase import CommandGroupBase


class ProbeCommandGroup(CommandGroupBase):
    """This command group contains functions for working with motorized prober.
    You are not meant to instantiate this class directly. Access it via the probe attribute
    of the [SentioProber](SentioProber.md) class.

    Example:

    ```py
    from sentio_prober_control.Sentio.ProberSentio import SentioProber

    prober = SentioProber.create_prober("tcpip", "127.0.0.1:35555")
    prober.probe.move_probe_xy(ProbeSentio.East, ProbeXYReference.Current, 1000, 2000)
    ```
    """

    def __init__(self, comm: CommunicatorBase) -> None:
        self.__comm = comm


    def async_step_probe_site(self, probe: ProbeSentio, idx: int) -> int:
        """Start the process of stepping to a positioner site.

        Each positioner can define n a number of predefined positions called "sites". This command
        initiates a step to such a site. This is an asynchronous command use in conjunction with
        ProberSentio.wait_complete() or ProberSentio.query_command_status().

        Args:
            probe: The probe to step.
            idx: The index of the site to step to.

        Returns:
            The async command id of the command.
        """

        self.__comm.send("start_step_positioner_site {0},{1}".format(probe.toSentioAbbr(), idx))
        resp = Response.check_resp(self.__comm.read_line())
        return resp.cmd_id()


    def async_step_probe_site_next(self, probe: ProbeSentio) -> int:
        """Step to next probe site.

        Each positioner can define n a number of predefined positions called "sites".
        This command initiates a step to the next site. This is an asynchronous command
        use in conjunction with ProberSentio.wait_complete() or ProberSentio.query_command_status().

        Args:
            probe: The probe to step.

        Returns:
            The async command id of the command.
        """

        self.__comm.send(
            "start_step_positioner_site_next {0}".format(probe.toSentioAbbr())
        )
        resp = Response.check_resp(self.__comm.read_line())
        return resp.cmd_id()


    def async_step_probe_site_first(self, probe: ProbeSentio) -> int:
        """Step to first probe site.

        Each positioner can define n a number of predefined positions called "sites".
        This command initiates a step to the first site. This is an asynchronous command
        use in conjunction with ProberSentio.wait_complete() or ProberSentio.query_command_status().

        Args:
            probe: The probe to step.

        Returns:
            The async command id of the command.
        """

        self.__comm.send(
            "start_step_positioner_site_first {0}".format(probe.toSentioAbbr())
        )
        resp = Response.check_resp(self.__comm.read_line())
        return resp.cmd_id()


    def get_probe_site(self, probe: ProbeSentio, idx: int) -> Tuple[int, str, float, float]:
        """Get information for a probe site.

        Each positioner can define n a number of predefined positions called "sites".
        This command returns the data associated with a site.

        Args:
            probe: The probe to get the site for.

        Returns:
            A tuple containing the site index, the site name, the x position in micrometer and the y position in micrometer.
        """
        self.__comm.send(
            "get_positioner_site {0},{1}".format(probe.toSentioAbbr(), idx)
        )
        resp = Response.check_resp(self.__comm.read_line())
        tok = resp.message().split(",")

        return int(tok[0]), str(tok[1]), float(tok[2]), float(tok[3])


    def get_probe_site_number(self, probe: ProbeSentio) -> int:
        """Returns the total number of probe sites set up.

        Each positioner can define n a number of predefined positions called "sites".
        This command returns the total number of sites.

        Returns:
            The total number of sites.
        """

        self.__comm.send("get_positioner_site_num {0}".format(probe.toSentioAbbr()))
        resp = Response.check_resp(self.__comm.read_line())
        return int(resp.message())


    def get_probe_xy(self, probe: ProbeSentio, ref: ProbeXYReference) -> Tuple[float, float]:
        """Get probe xy position in micrometer.

        Args:
            probe: The probe to get the position for.
            ref: The position reference for the returned values.

        Returns:
            A tuple containing the x and y position in micrometer.
        """

        self.__comm.send(
            "get_positioner_xy {0},{1}".format(probe.toSentioAbbr(), ref.toSentioAbbr())
        )
        resp = Response.check_resp(self.__comm.read_line())
        tok = resp.message().split(",")
        return float(tok[0]), float(tok[1])


    def get_probe_z(self, probe: ProbeSentio, ref: ProbeZReference) -> float:
        """Get probe z position in micrometer.

        Args:
            probe: The probe to get the position for.
            ref: The position reference for the returned values.

        Returns:
            The z position in micrometer.
        """
        self.__comm.send(
            "get_positioner_z {0}, {1}".format(probe.toSentioAbbr(), ref.toSentioAbbr())
        )
        resp = Response.check_resp(self.__comm.read_line())
        return float(resp.message())


    def move_probe_contact(self, probe: ProbeSentio) -> float:
        """Move a probe to its contact position.

        Args:
            probe: The probe to move.

        Returns:
            The z position after the move in micrometer (from zero).
        """

        self.__comm.send("move_positioner_contact {0}".format(probe.toSentioAbbr()))
        resp = Response.check_resp(self.__comm.read_line())
        return float(resp.message())


    def move_probe_separation(self, probe: ProbeSentio) -> float:
        """Move a probe to its separation position.

        Args:
            probe: The probe to move.

        Returns:
            The z position after the move in micrometer (from zero).
        """

        self.__comm.send("move_positioner_separation {0}".format(probe.toSentioAbbr()))
        resp = Response.check_resp(self.__comm.read_line())
        return float(resp.message())


    def move_probe_home(self, probe: ProbeSentio) -> Tuple[float, float]:
        """Move probe to its home position.

        Args:
            probe: The probe to move.

        Returns:
            A tuple containing the x and y position after the move.
        """

        self.__comm.send("move_positioner_home {0}".format(probe.toSentioAbbr()))
        resp = Response.check_resp(self.__comm.read_line())
        tok = resp.message().split(",")
        return float(tok[0]), float(tok[1])


    def move_probe_xy(self, probe: ProbeSentio, ref: ProbeXYReference, x: float, y: float) -> Tuple[float, float]:
        """Move probe to a given position.

        Args:
            probe: The probe to move.
            ref: The position reference for the submitted values.
            x: The x position in micrometer.
            y: The y position in micrometer.

        Returns:
            A tuple containing the x and y position after the move.
        """

        self.__comm.send(f"move_positioner_xy {probe.toSentioAbbr()},{ref.toSentioAbbr()},{x},{y}")
        resp = Response.check_resp(self.__comm.read_line())
        tok = resp.message().split(",")
        return float(tok[0]), float(tok[1])


    def move_probe_z(self, probe: ProbeSentio, ref: ProbeZReference, z: float) -> float:
        """Move probe to a given z position.

        Args:
            probe: The probe to move.
            ref: The position reference for the submitted values.
            z: The target z position in micrometer.

        Returns:
            The z position after the move in micrometer (from zero).
        """

        self.__comm.send(f"move_positioner_z {probe.toSentioAbbr()}, {ref.toSentioAbbr()}, {z}")
        resp = Response.check_resp(self.__comm.read_line())
        return float(resp.message())


    def set_probe_contact(self, probe: ProbeSentio, z: float | None = None) -> None:
        """Set contact position of a positioner.

        Args:
            probe: The probe to set contact height.
            z: The contact height in micrometer. If not specified, the current z position is used.
        """

        if z == None:
            self.__comm.send("set_positioner_contact {0}".format(probe.toSentioAbbr()))
        else:
            self.__comm.send("set_positioner_contact {0},{1}".format(probe.toSentioAbbr(), z))

        Response.check_resp(self.__comm.read_line())


    def set_probe_home(self, probe: ProbeSentio, site: ChuckSite | None = None, x: float | None = None, y: float | None = None) -> None:
        """Set home position of a probe.

        Args:
            probe: The probe to set home position.
            site: The chuck site to set the home position for. If None is specified the current site is used.
            x: The x position in micrometer. If not specified, the current x position is used.
            y: The y position in micrometer. If not specified, the current y position is used.
        """
        if site is None:
            self.__comm.send(f"set_positioner_home {probe.toSentioAbbr()}")
        else:
            self.__comm.send(f"set_positioner_home {probe.toSentioAbbr()}, {site.toSentioAbbr()}, {x}, {y}")

        Response.check_resp(self.__comm.read_line())


    def step_probe_site(self, probe: ProbeSentio, idx: int) -> Tuple[str, float, float]:
        """Step to a specific probe site.

        Each positioner can define n a number of predefined positions called "sites".
        This command initiates a step to such a site.

        Args:
            probe: The probe to step.
            idx: The index of the site to step to.

        Returns:
            A tuple containing the site id, the x position in micrometer and the y position in micrometer.
        """

        self.__comm.send("step_positioner_site {0},{1}".format(probe.toSentioAbbr(), idx))

        resp = Response.check_resp(self.__comm.read_line())
        tok = resp.message().split(",")
        return tok[0], float(tok[1]), float(tok[2])


    def step_probe_site_first(self, probe: ProbeSentio) -> Tuple[str, float, float]:
        """Step to the first probe site.

        Each positioner can define n a number of predefined positions called "sites".
        This command initiates a step to the first site.

        Args:
            probe: The probe to step.

        Returns:
            A tuple containing the site id, the x position in micrometer and the y position in micrometer.
        """

        self.__comm.send("step_positioner_site_first {0}".format(probe.toSentioAbbr()))
        resp = Response.check_resp(self.__comm.read_line())
        tok = resp.message().split(",")
        return tok[0], float(tok[1]), float(tok[2])


    def step_probe_site_next(self, probe: ProbeSentio) -> Tuple[str, float, float]:
        """Step to the next probe site.

        Each positioner can define n a number of predefined positions called "sites".
        This command initiates a step to the next site.

        Args:
            probe: The probe to step.

        Returns:
            A tuple containing the site id, the x position in micrometer and the y position in micrometer.
        """

        self.__comm.send("step_positioner_site_next {0}".format(probe.toSentioAbbr()))
        resp = Response.check_resp(self.__comm.read_line())
        tok = resp.message().split(",")
        return tok[0], float(tok[1]), float(tok[2])



""" # Communication package
    
    This package contains classes for communication with the MPI probe station with
    different protocols: TCP/IP, Gpib or VISA. Communication is handled by instances of Communicator objects.
    When initializing a probe station a communicator object must be passed to the constructor of the probe 
    station but the communicator could also be used standalone.

    # Communication Methods
    * `sentio_prober_control.Communication.CommunicatorTcpIp`<br/>for plain TCP/IP communication
    * `sentio_prober_control.Communication.CommunicatorGpib`<br/>for GPIB communication via native drivers (ADLINK and NI)
    * `sentio_prober_control.Communication.CommunicatorVisa`<br/>for using the NI-VISA interface which warps (TCP/IP, GPIB and RS232) 

    ## Communicator Example

    The following example will establish a Tcp/Ip connection to a locally running SENTIO instance. It will then send a "*IDN?" command
    in order to read the system information of the probe station back.

    >>> from sentio_prober_control.Communication.CommunicatorTcpIp import CommunicatorTcpIp
    >>> comm = CommunicatorTcpIp.create("127.0.0.1:35555")
    >>> comm.send("*IDN?")
    >>> comm.read_line()
    'MPI Corporation,Sentio - Probe System Software Suite,0,23.2.99.0'
    
"""
print(f"Initializing package {__name__}")



from deprecated import deprecated

from sentio_prober_control.Sentio.Response import Response
from sentio_prober_control.Sentio.Enumerations import Compensation, ExecuteCompensation, OnTheFlyMode
from sentio_prober_control.Sentio.CommandGroups.CommandGroupBase import CommandGroupBase


@deprecated(reason="duplicated; This command group is already a subgroup of the vision command group.")
class CompensationCommandGroup(CommandGroupBase):
    """This command group contains functions for working with x,y and z compensation.

    This command group was marked deprecated on 2023-09-04. It will be removed in a future version.
    Use the SentioProber.vis.compensation group instead!
    """

    def __init__(self, comm) -> None:
        """@private"""
        super().__init__(comm)


    @deprecated(reason="Use prober.vis.compensation.set_compensation instead.")
    def set_compensation(self, comp: Compensation, enable: bool):
        """Enable or disable compensation for a given subsystem.

        This command was marked deprecated on 2023-09-04. It will be removed in a future version.
        Use the command in the visioncompensation group instead:

        >>> prober.vis.compensation.set_compensation(comp, enable)
        """

        self.comm.send(f"vis:compensation:enable {comp.toSentioAbbr()}, {enable}")
        resp = Response.check_resp(self.comm.read_line())
        tok = resp.message().split(",")
        return tok[0], tok[1]


    @deprecated(reason="Use prober.vis.compensation.start_execute instead.")
    def execute_compensation(self, comp: ExecuteCompensation, mode: OnTheFlyMode):
        """Execute compensation.

        This command was marked deprecated on 2023-09-04. It will be removed in a future version.
        use prober.vis.compensation.start_execute instead.

        """
        self.comm.send(
            "vis:compensation:start_execute {},{}".format(
                comp.toSentioAbbr(), mode.toSentioAbbr()
            )
        )
        resp = Response.check_resp(self.comm.read_line())
        return resp



from typing import Tuple

from sentio_prober_control.Sentio.Enumerations import AxisOrient, StatusBits
from sentio_prober_control.Sentio.Response import Response
from sentio_prober_control.Sentio.CommandGroups.CommandGroupBase import CommandGroupBase


class WafermapSubsiteGroup(CommandGroupBase):
    """Represents the wafermap subsite command group which provides
    functionality for setting up und stepping over subsites.
    """

    def __init__(self, comm, wafermap_command_group) -> None:
        """Creates a new WafermapSubsiteGroup object.

        You are not meant to directly create objects of this class.
        """
        super().__init__(comm)
        self._parent_command_group = wafermap_command_group


    def add(self, id: str, x: float, y: float, orient: AxisOrient = AxisOrient.UpRight) -> None:
        """Add a single subsite to the wafermap.

        Creates a new subsite definition in SENTIO. The subsite position is defined
        relative to the die home position. This is the chuck home position projected onto
        the current die.

        Wraps the "map:subsite:add" remote command.

        Args:
            id: The subsite id.
            x: The x position of the subsite in micrometer as an offset to the die home position.
            y: The y position of the subsite in micrometer as an offset to the die home position.
            orient: The axis orientation used fot the submitted values
        """
        self.comm.send("map:subsite:add {}, {}, {}, {}".format(id, x, y, orient.toSentioAbbr()))
        Response.check_resp(self.comm.read_line())


    def bin_step_next(self, bin: int) -> Tuple[int, int, int]:
        """Step to the next active subsite and assign bin code to current subsite.

        Wraps the "map:subsite:bin_step_next" remote command.

        Args:
            bin: The bin code to assign to the current subsite.

        Returns:
            A tuple containing the wafermap row, column and subsite index after the step.
        """
        self.comm.send(f"map:subsite:bin_step_next {bin}")

        resp = Response.check_resp(self.comm.read_line())
        self._parent_command_group.__end_of_route = (resp.status() & StatusBits.EndOfRoute) == StatusBits.EndOfRoute

        tok = resp.message().split(",")
        return int(tok[0]), int(tok[1]), int(tok[2])


    def get(self, idx: int, orient: AxisOrient | None = None) -> Tuple[str, float, float]:
        """Returns the subsite definition for a subsite with a given index.

        Wraps the "map:subsite:get" remote command.

        Args:
            idx: The index of the subsite.
            orient: The axis orientation used for the returned subsite coordinates. If this parameter is omitted the axis orientation of the wafer map is used.

        Returns:
            A tuple containing the subsite id, the x position and the y position of the subsite.
            X and y positions are relative to the die home position.
        """

        # If no orientation is given use the map axis orientation. The SENTIO remote command would always default to UpRight which may 
        # not be correct if the map has a different orientation.
        if orient is None:
            orient_str = "MAP"
        else:
            orient_str = orient.toSentioAbbr()

        self.comm.send(f"map:subsite:get {idx}, {orient_str}")
        resp = Response.check_resp(self.comm.read_line())

        tok = resp.message().split(",")
        return str(tok[0]), float(tok[1]), float(tok[2])


    def get_num(self) -> int:
        """Retrieve the number of subsites per die defined in the wafermap.

        Wraps the "map:subsite:get_num" remote command.

        Returns:
            The number of subsites in the wafermap.
        """
        self.comm.send("map:subsite:get_num")
        resp = Response.check_resp(self.comm.read_line())
        return int(resp.message())


    def reset(self) -> None:
        """Reset Sentios subsite definitions.

        Wraps the "map:subsite:reset" remote command.
        """
        self.comm.send("map:subsite:reset")
        Response.check_resp(self.comm.read_line())


    def step_next(self) -> Tuple[int, int, int]:
        """Step to the next active subsite.

        Wraps the "map:subsite:step_next" remote command.

        Returns:
            A tuple containing the wafermap row, column and subsite index after the step.
        """
        self.comm.send("map:subsite:step_next")

        resp = Response.check_resp(self.comm.read_line())
        self._parent_command_group.__end_of_route = (resp.status() & StatusBits.EndOfRoute) == StatusBits.EndOfRoute

        tok = resp.message().split(",")
        return int(tok[0]), int(tok[1]), int(tok[2])



from enum import Enum

from sentio_prober_control.Communication.CommunicatorBase import CommunicatorBase
from sentio_prober_control.Devices.GpibAdlinkDriver import GpibAdlinkDriver
from sentio_prober_control.Devices.GpibNiDriver import GpibNiDriver


class GpibCardVendor(Enum):
    """An enumeration that specifies the type of the GPIB card to use for GPIB communication."""

    Adlink = (0,)
    """ Specify this if you use an ADLINK GPIB card. You must have installed the ADLINK drivers on your system in order to use this option! """

    NationalInstruments = 1
    """ Specify this if you use a National Instruments GPIB card. You must have installed the NI-GPIB drivers on your systen in order to use this option! """


class CommunicatorGpib(CommunicatorBase):
    """Communicator for GPIB communication.

    This class is a wrapper around the native GPIB drivers. You can
    use it either with ADLINK or National Instruments GPIB cards.
    The native drivers must be installed on the system or this class
    will not work.
    """

    def __init__(self, vendor: GpibCardVendor):
        """Construcst a GPIB communicator.

        :param vendor: Specifies the native driver to use (either Adlink or NI).
        """
        if vendor == GpibCardVendor.Adlink:
            self._driver = GpibAdlinkDriver()
        elif vendor == GpibCardVendor.NationalInstruments:
            self._driver = GpibNiDriver()
        else:
            raise NotImplementedError(
                "Driver for Gpib card vendor {0}not implemented!".format(vendor)
            )

    @staticmethod
    def create(vendor: GpibCardVendor, addr: str):
        """Create an instance of a GPIB communicator.

        This functions creates an instance of a GPIB communicator and connects it to a given address.

        :param vendor: Specifies the native driver to use (either Adlink or NI)
        :param addr: A string that specifies the address of the GPIB device to connect to. The address must have the format "BOARD_NAME&colon;ADDRESS"
        :return: returns an instance of a GPIB communicator.
        :raises Exception: If the address is empty or does not have the correct format.
        """
        c = CommunicatorGpib(vendor)
        c.connect(addr)
        return c

    def connect(self, address: str):
        """Connects to a GPIB device at the specified address.

        :param addr: A string that specifies the address of the GPIB device to connect to. The address must have the format "BOARD_NAME&colon;ADDRESS"
        :raises Exception: If the address is empty or does not have the correct format.
        """
        if address is None:
            raise Exception("Gpib address must not be empty!")

        tok = address.split(":")
        if len(tok) != 2:
            raise Exception('Gpib address must have the format "BOARD_NAME:ADDRESS"')

        board: str = tok[0]
        address: int = int(tok[1])
        self._driver.connect(board, address)

    def disconnect(self):
        """Disconnects from the GPIB device."""
        pass

    def send(self, msg: str):
        """Send a text string via the communication interface.

        :param msg: The text string to send.
        """
        self._driver.send(msg)

    def read_line(self):
        """Read a line from the communication interface."""
        return self._driver.receive().rstrip()
